#!/usr/bin/perl -w
# $Id: sys-backup.pl 1 2007-06-04 12:47:41Z rea $
#
# Backup utility that performs the backups for the specified system
# files. File carrying the list of files (one file -- one line, lines
# starting with '#' are comments, empty lines are skipped) should be
# specified on the command line.
#
# The supplemental hashfile will be used: all its lines have the
# filename and hash pairs delimited by $delimiter parameter.
#

use strict;
use File::Copy;
use Digest::SHA;

sub usage(;$);
sub deepcopy($$);
sub fixperms($$);
sub getperms($);
sub fullpath($$);
sub createpath($$);

my ($filename, $hashfile, $name, $hash, %names, %hashes, $base);
my ($src, $dst, %tocopy, @todelete, $i);

my ($delimiter, $hashalg, $sha, $curdigest, $olddigest, $verbose);
$delimiter = "\t";
$hashalg = "sha256";
$sha = Digest::SHA->new($hashalg);
if (!defined($sha)) {
	print STDERR "Unable to instantiate '" . $hashalg .
	    "' hash. Exiting.\n";
	exit(255);
}
$verbose = 1;

if (!defined($ARGV[0])) {
	usage("missing first argument");
	exit(255);
}
$filename = $ARGV[0];
$hashfile = $filename . ".hash";

if (!defined($ARGV[1])) {
	usage("missing second argument");
	exit(255);
}
$base = $ARGV[1];
if (!-d $base) {
	usage("second argument '" . $base . "' is not a directory");
	exit(255);
}
# Add the trailing slash: we are depend on it.
if (substr($base, -1, 1) ne "/") {
	$base .= "/";
}

# Reading filenames and the current hashes.
if (!-f $filename) {
	usage("'" . $filename . "' is not a file.");
	exit(255);
}
if (!open(LIST, $filename)) {
	usage("Unable to open '" . $filename . "' for reading: " . $!);
	exit(255);
}
while (<LIST>) {
	chomp;
	if (/^#/) { next; }
	if (substr($_, 0, 1) ne "/") {
		print STDERR "Not an absolute path '" . $_ . "' .
		    Skipping it.\n";
		next;
	}
	$names{$_} = 1;
}
close(LIST);
if (open(HASH, $hashfile)) {
	while (<HASH>) {
		chomp;
		($name, $hash) = split(/$delimiter/);
		$hashes{$name} = $hash;
	}
	close(HASH);
} else {
	print STDERR <<"EOF";
Unable to open '$hashfile' for reading.
Will copy all files and regenerate hashes.
EOF
}

#
# Create directories (and synchronize permissions) for all specified
# files. This is done for all files, since directory permissions
# could be changed since the previous run even for the unchanged
# files.
#
foreach $src (keys %names) {
	createpath($src, $base);
}

#
# Determining new and changed files. The 'hashes' hash is updated
# with new hashes and new files (with their hashes).
#
foreach $src (keys %names) {
	$dst = fullpath($base, $src);
	if (!-r $src) {
		print STDERR "File '" . $src .
		    "' is unreadable to me. Skipping it.\n";
		next;
	}
	$sha->reset();
	$sha->addfile($src);
	$curdigest = $sha->hexdigest;
	$olddigest = $hashes{$src};
	if (defined($olddigest)) {
		if ($curdigest ne $olddigest || !-r $dst) {
			$tocopy{$src} = $curdigest;
		}
	} else {
		$tocopy{$src} = $curdigest;
	}
}

#
# Determining deleted files. The names of the deleted files are
# removed from the 'hashes' hash.
#
foreach $src (keys %hashes) {
	if (!defined($names{$src})) {
		$todelete[$#todelete + 1] = $src;
		delete $hashes{$src};
	}
}

if (!open(HASH, ">", $hashfile)) {
	usage("Unable to open '" . $hashfile . "' for writing. Exiting.");
	exit(255);
}

#
# Copying changed files, writing their hashes to the hashfile
# and deleting the 'hashes' entries for each successfully
# processed file. Only unprocessed files (both untouched and with
# failed copying) will be left in the 'hashes' after the block.
#
if ($verbose && scalar(keys %tocopy) != 0) {
	print "Files to be copied (with their hashes):\n";
}
foreach $src (keys %tocopy) {
	my ($error);

	$dst = $base . $src;
	if ($verbose) {
		print $src . $delimiter . $tocopy{$src} . "\n";
	}
	$error = deepcopy($src, $base);
	if (!defined($error)) {
		print HASH $src . $delimiter . $tocopy{$src} . "\n";
		delete $hashes{$src};
	} else {
		print STDERR "Copy failed: '" . $src . "' => '" . $dst .
		    "': " . $error . "\n";
	}
}

if ($verbose && scalar(@todelete)) {
	print "Files to be deleted:\n";
}
for ($i = 0; $i < scalar(@todelete); $i++) {
	$src = $todelete[$i];
	if ($verbose) {
		print $src . "\n";
	}
}

#
# Dump all entries left in the 'hashes' hash: unmodified files
# or those for which copy process was failed.
#
foreach $src (keys %hashes) {
	print HASH $src . $delimiter . $hashes{$src} . "\n";
}

close(HASH);

exit(0);

#
# Prints usage statement. If the first argument is given, it is
# printed before the usage preceeded by the string 'Error: '
# and terminated by newline.
#
sub usage(;$)
{
	if (defined($_[0])) {
		print STDERR "Error: " . $_[0] . "\n\n";
	}
	print STDERR <<"EOF";
Usage: $0 filelist basedir

File 'filelist' carries the filenames to backup. One line -- one
file name, lines starting with '#' symbol and empty lines are
skipped. File names must be absolute.

'basedir' is the directory where the files will be copied to. It
is the root of the tree: the path of each original file will be
appended to the 'basedir'.
EOF
}

#
# Copies the file $_[0] to the directory $_[1] using it as the
# base directory: the final file name will be the concatenation
# of the $_[0] and $_[1].
#
# We assume that the target directory already exists: it should
# be prepared by the 'createpath' routine.
#
# Returns 'undef' if all went OK and the error description if
# something went wrong.
#
sub deepcopy($$) {
	my ($src, $dst, $basedir, $perm, $umask, $err);

	($src, $basedir) = @_;
	if (substr($src, 0, 1) ne "/") {
		return "source '" . $src . "' is not an absolute path";
	}
	if (substr($basedir, 0, 1) ne "/") {
		return "base directory '" . $basedir .
		    "' is not an absolute path";
	}
	if (!-d $basedir) {
		return "base directory argument '" . $basedir .
		    "' is not a directory";
	}
	if (substr($src, -1, 1) eq "/") {
		return "source '" . $src . "' is a directory path";
	}
	$dst = fullpath($basedir, $src);

	$perm = getperms($src);
	if ($! != 0) {
		return "unable to get permissions for '" . $src . "': " . $!;
	}

	$umask = umask(0077);
	if (!copy($src, $dst)) {
		umask($umask);
		return $!;
	}
	umask($umask);

	$err = fixperms($perm, $dst);
	if (defined($err)) {
		unlink($dst);
		return $err;
	}

	return undef;
}

#
# Get permissions for the specified object. Sets '$!'.
# Returns the permissions.
#
sub getperms($) {
	my ($src, $perm, $err);
	$src = shift;

	$perm = (stat($src))[2];
	$err = $!;
	if (!defined($perm)) {
		$! = $err;
		return 0;
	}
	$perm &= 07777;

	$! = 0;
	return $perm;
}

#
# Sets the permissions given by the first argument on the filesystem
# object given by the second argument. Returns 'undef' for the
# success or the the error message.
#
sub fixperms($$) {
	my ($perm, $dst);

	($perm, $dst) = @_;

	if (!chmod($perm, $dst)) {
		return sprintf("chmod(%04o) failed: ", $perm) . $!;
	}
	return undef;
}

#
# Creates the full name of the file $_[1] rooted at the directory
# $_[0]. Just concatenates the $_[0] and $_[1]. $_[0] assumed
# to be slash-ended.
#
sub fullpath($$)
{
	return $_[0] . $_[1];
}

#
# Creates the path for the file $_[0] under the directory $_[1]
# and synchronizes the permissions of all patch components.
#
# Returns 'undef' if all is OK, and the error description
# otherwise.
#
sub createpath($$) {
	my ($src, $dst, $basedir, @path, $i, $err, $hier, $dir, $perm);

	($src, $basedir) = @_;

	# $path[0] will be empty since we check for the absolute path
	@path = split('/', $src);
	$hier = "";
	for ($i = 1; $i < scalar(@path) - 1; $i++) {
		$hier .= "/" . $path[$i];
		$dir = fullpath($basedir, $hier);
		if (!-d $dir && !mkdir $dir) {
			return "unable to create directory '" . $dir .
			    "': " . $!;
		}
		$perm = getperms($hier);
		if ($! != 0) {
			return "unable to obtain permissions for '" . $hier .
			    "': " . $!;
		}
		$err = fixperms($perm, $dir);
		if (defined($err)) {
			return $err;
		}
	}
	return undef;
}
