/*
 * Eygene Ryabinkin, rea@codelabs.ru.  Copyright 2009.
 *
 * Helper utility to make the list of file and directory permissions
 * for the specified filesystem objects.  Input list is read from the
 * standard input, line format is "path [mode] [group:owner]", comments
 * are denoted by '#' on the first character of the line.  Leading whitespace
 * is stripped.  Comments and blank lines are piped through unmodified.
 * 'path' strings must start with '/'.
 *
 * We will output the resulting list to the standard output.  I will try
 * to preserve the whitespace structure for the entries that have 'mode'
 * and/or 'group:owner' components.  Entries that have only 'path'
 * component will be printf-formatted with "%-50s  %s  %s\n".
 *
 * Utility has two modes:
 * - all: 'mode' and 'group:owner' are checked (and modified) for
 *   each entry;
 * - missing: 'mode' and 'group:owner' are added only for entries
 *   that are missing any of these entries.
 *
 */

#include "config.h"

#ifdef LIMITS_H_IS_IN_SYS
#include <sys/limits.h>
#else
#include <limits.h>
#endif

#include <sys/types.h>
#include <sys/stat.h>

#include <assert.h>
#include <ctype.h>
#include <errno.h>
#include <grp.h>
#include <pwd.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "strbuf.h"
#include "err.h"
#include "sysexits.h"
#include "xalloc.h"

enum utility_mode {
	mode_all,		/**< Adds all current permissions. */
	mode_missing		/**< Adds only missing permissions. */
};

void
usage(const char *_argv0);
void
process_flist_line(char *_line, FILE *_fp, enum utility_mode _m);
void
parse_flist_line(char *_line, char **_name, size_t _name_extra,
  unsigned long *_mode, char **_user, char **_group);
int
map_uid_gid(uid_t _uid, gid_t _gid, char **_user, char **_group);

int
main(int argc, char *argv[])
{
	int in_fd = 0;
	FILE *out_fp = stdout;
	struct strbuf *buf = NULL;
	enum utility_mode mode;

	if (argc < 2) {
		usage(argv[0]);
		exit(EX_USAGE);
	}

	if (strcmp(argv[1], "all") == 0) {
		mode = mode_all;
	} else if (strcmp(argv[1], "missing") == 0) {
		mode = mode_missing;
	} else {
		fprintf(stderr, "Unknown utility mode '%s'.\n",
		  argv[1]);
		exit(EX_USAGE);
	}

	buf = strbuf_create(80);
	if (buf == NULL)
		err(EX_OSERR, "Unable to allocate string buffer");

	strbuf_errno_clear(buf);
	strbuf_set_fd(buf, in_fd);
	while (1) {
		char *line = strbuf_readline(buf);
		size_t i;

		if (line == NULL) {
			if (strbuf_errno_get(buf) == strbuf_e_eof)
				break;
			else
				errx(EX_OSERR,
				  "Can not read input line: %s",
				  strbuf_errno_to_s(buf));
		}

		for (i = 0; isspace(line[i]); i++);
		if (line[i] == '#' || line[i] == '\0') {
			fputs(line, out_fp);
			continue;
		}
		process_flist_line(line, out_fp, mode);
	}

	strbuf_destroy(buf);

	return 0;
}


/**
 * \brief Prints the usage statement.
 *
 * \param argv0 name of the program.
 */
void
usage(const char *argv0)
{
	printf("Usage: %s [all|missing]\n", argv0);
}


/**
 * \brief Processes the line that describes a filesystem
 * object.  Depending on the operation mode, this function
 * will:
 * - (mode_all): add or modify all ownership and permission
 *   information;
 * - (mode_missing): add only missing ownership and/or
 *   permission information.
 *
 * \param line line to process.
 * \param fp here we will write the processed line to.
 * \param m mode in which we will operate.
 */
void
process_flist_line(char *line, FILE *fp, enum utility_mode m)
{
	char *name = NULL;
	unsigned long mode = ULONG_MAX;
	char *user = NULL, *group = NULL;
	unsigned long a_mode = ULONG_MAX;
	char *a_user = NULL, *a_group = NULL;
	struct stat sb;

	assert(line != NULL);
	assert(fp != NULL);

	parse_flist_line(line, &name, 1, &mode, &user, &group);
	if (name == NULL) {
		/* Print this unparseable line as-is. */
		fputs(line, fp);
		goto _bailout;
	}
	if (lstat(name, &sb)) {
		/* No stat -- comment this line out. */
		fputs("#NONEXISTENT ", fp);
		fputs(line, fp);
		goto _bailout;
	}

	/* Add trailing slash to the directory names if needed */
	if (S_ISDIR(sb.st_mode)) {
		size_t len = strlen(name);
		unsigned char slash =
		  (len > 0 && name[len - 1] == '/');
		if (!slash) strcat(name, "/");
	}

	if (m == mode_missing && mode != ULONG_MAX &&
	    user != NULL && group != NULL) {
		/* We're done here. */
		fputs(line, fp);
		goto _bailout;
	}

	/* Collect current metadata. */
	a_mode = sb.st_mode & (S_IRWXU|S_IRWXG|S_IRWXO);
	if (map_uid_gid(sb.st_uid, sb.st_gid, &a_user, &a_group)) {
		/*
		 * Can't get actual ownership, comment out this entry.
		 */
		fputs("#NO METADATA ", fp);
		fputs(line, fp);
		goto _bailout;
	}

	if (mode == a_mode && strcmp(user, a_user) == 0 &&
	    strcmp(group, a_group) == 0) {
		/* Just reproduce the string in its virgin form. */
		fputs(line, fp);
		goto _bailout;
	}

	fprintf(fp, "%-50s  %04o  %s:%s\n", name,
	  (unsigned)(m == mode_all || mode == ULONG_MAX ? a_mode : mode),
	  (m == mode_all || user == NULL ? a_user : user),
	  (m == mode_all || group == NULL ? a_group : group));
_bailout:
	if (name != NULL)
		xfree(name);
	if (user != NULL)
		xfree(user);
	if (group != NULL)
		xfree(group);
	if (a_user != NULL)
		xfree(a_user);
	if (a_group != NULL)
		xfree(a_group);
}


/**
 * \brief Parses the line from the filelist
 * and decomposes it to the filename, mode,
 * owner and group.
 *
 * \param line string to parse; it should be
 * write-able, because we will parse it in-place
 * and insert some '\0' characters into this string.
 * \param name pointer to place the filename to.
 * \param name_extra extra bytes to add to the name
 *        buffer size: enables one to put some more
 *        characters after the name.
 * \param mode pointer to place the file mode to.
 * \param user pointer to place the user name to.
 * \param group pointer to place the group name to.
 *
 * All string pointers will be initialized to NULL at the
 * beginning; mode will be initialized to ULONG_MAX.
 * If any returned pointer will contain this value,
 * it means that the corresponding parameter wasn't
 * correctly parsed.  Non-NULL pointer to string means
 * that this memory was allocated wia *alloc(), so it
 * must be freed after use.
 */
void
parse_flist_line(char *line, char **name, size_t name_extra,
  unsigned long *mode, char **user, char **group)
{
	size_t i, len;
	char *p = NULL;

	assert(line != NULL);
	assert(name != NULL);
	assert(mode != NULL);
	assert(user != NULL);
	assert(group != NULL);

	*name = NULL;
	*mode = ULONG_MAX;
	*user = NULL;
	*group = NULL;

	/* Filename */
	for (i = 0; isspace(line[i]); i++);
	p = line + i;
	for (; line[i] && !isspace(line[i]); i++);
	len = line + i - p;
	if (len == 0)
		return;
	*name = xmalloc(len + 1 + name_extra);
	memcpy(*name, p, len);;
	(*name)[len] = '\0';
	if (line[i] == '\0')
		return;
	else
		i++;

	/* Octal mode */
	for (; isspace(line[i]); i++);
	p = line + i;
	for (; line[i] && line[i] >= '0' && line[i] <= '7'; i++);
	if (p == line + i)
		return;
	/*
	 * XXX: Not terminating the string here: strtoul() should
	 * XXX: stop on the first non-octal character of the end
	 * XXX: of line, just as we had been doing above.
	 */
	errno = 0;
	*mode = strtoul(p, NULL, 8);
	if (errno) {
		*mode = 0;
		return;
	}
	if (line[i] == '\0')
		return;
	else
		i++;

	/* user:group */
	for (; isspace(line[i]); i++);
	p = line + i;
	for (; line[i] && line[i] != ':'; i++);
	len = line + i - p;
	if (len == 0)
		return;
	*user = xmalloc(len + 1);
	memcpy(*user, p, len);
	(*user)[len] = '\0';
	if (line[i] == '\0')
		return;
	else
		i++;
	p = line + i;
	for (; line[i] && !isspace(line[i]); i++);
	len = line + i - p;
	if (len == 0)
		return;
	*group = xmalloc(len + 1);
	memcpy(*group, p, len);
	(*group)[len] = '\0';
	if (line[i] == '\0')
		return;
	else
		i++;
}


/**
 * \brief Maps UID and GID the names of owner and group.
 *
 * If we can't properly map the UID or GID into name,
 * we will just emit UID or GID in their numerical representation.
 *
 * \param uid UID to translate
 * \param gid GID to translate
 * \param user pointer to place the user name to
 * \param group pointer to place the user name to
 *
 * If any returned pointer is not NULL, it must be released with
 * xfree() once it is no longer needed.
 *
 * \return 0 if everything went well and -1 for the error in
 * translation of any parameter.
 */
int
map_uid_gid(uid_t uid, gid_t gid, char **user, char **group)
{
	int len;
	struct passwd *pw;
	struct group *gr;

	assert(user != NULL);
	assert(group != NULL);

	*user = NULL;
	*group = NULL;

	pw = getpwuid(uid);
	if (pw == NULL || pw->pw_name == NULL) {
		len = snprintf(NULL, 0, "%ld", (long)uid);
		if (len < 0)
			return -1;
		*user = xmalloc(len + 1);
		snprintf(*user, len + 1, "%ld", (long)uid);
	} else {
		len = strlen(pw->pw_name);
		*user = xmalloc(len + 1);
		strncpy(*user, pw->pw_name, len + 1);
	}

	gr = getgrgid(gid);
	if (gr == NULL || gr->gr_name == NULL) {
		len = snprintf(NULL, 0, "%ld", (long)gid);
		if (len < 0) {
			if (*user != NULL) {
				xfree(*user);
				*user = NULL;
			}
			return -1;
		}
		*group = xmalloc(len + 1);
		snprintf(*group, len + 1, "%ld", (long)gid);
	} else {
		len = strlen(gr->gr_name);
		*group = xmalloc(len + 1);
		strncpy(*group, gr->gr_name, len + 1);
	}

	return 0;
}
