/**
 * \brief Small library that implements line reader.
 *
 * It tries to read the entire line from the file into the
 * dynamically-sized buffer.
 *
 * Eygene Ryabinkin, rea@codelabs.ru.  Copyright 2009, 2011.
 */

#include "config.h"

#ifdef LIMITS_H_IS_IN_SYS
#include <sys/limits.h>
#else
#include <limits.h>
#endif

#include <sys/types.h>

#include <assert.h>
#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <strings.h>
#include <unistd.h>

#include "strbuf.h"

/* Size for the readahead buffer */
#define READAHEAD_SIZE		2048

/**
 * Main structure that holds the state information
 * for the current string buffer.
 */
struct strbuf {
	char *buf;		/**< Buffer for the string */
	size_t size;		/**< Buffer size */
	int fd;			/**< Current file descriptor */
	off_t offset;		/**< Current offset in the buffer */
	char *readahead;	/**< Readahead buffer */
	char *ra_pos;		/**< Position in the readahead buffer */
	size_t ra_unread;	/**< Unread characters in readahead buffer */
	enum strbuf_errno_t lasterr;	/**< Last error */
};

/*
 * Each function will set the lasterr field if it was failed.
 * Successful invocations won't alter lasterr, so it shouldn't
 * be examined for the successful calls.
 */

static int
_strbuf_grow(struct strbuf *_s);
static void
_strbuf_make_empty(struct strbuf *_s);
static int
_strbuf_finalize_string(struct strbuf *_s);
static int
_strbuf_add_char(struct strbuf *_s, char _c);
static ssize_t
_strbuf_fill_buffer(struct strbuf *_s);


/**
 * \brief Creates strbuf with the given initial size.
 *
 * \param initial initial size for the strbuf.
 *
 * \return strbuf descriptor (pointer to a describing structure)
 * or NULL if creation failed.
 */
struct strbuf *
strbuf_create(size_t initial)
{
	struct strbuf *s = NULL;

	if (initial == 0)
		initial = 16;

	s = (struct strbuf *)malloc(sizeof(*s));
	if (s == NULL) {
		errno = ENOMEM;
		return NULL;
	}
	bzero((void *)s, sizeof(*s));

	s->buf = (char *)calloc(initial, sizeof(s->buf[0]));
	if (s->buf == NULL) {
		free((void *)s);
		errno = ENOMEM;
		return NULL;
	}
	bzero((void *)s->buf, initial * sizeof(s->buf[0]));
	s->size = initial;

	s->fd = -1;

	s->readahead =
	  (char *)calloc(READAHEAD_SIZE, sizeof(s->readahead[0]));
	if (s->readahead == NULL) {
		free((void *)s->buf);
		free((void *)s);
		errno = ENOMEM;
		return NULL;
	}

	return s;
}


/**
 * \brief Destroys strbuf and associated memory.
 *
 * \param s pointer to the strbuf structure to destroy.
 */
void
strbuf_destroy(struct strbuf *s)
{
	assert(s != NULL);

	if (s->buf) {
		free((void *)s->buf);
		s->buf = NULL;
	}
	if (s->readahead) {
		free((void *)s->readahead);
		s->readahead = NULL;
	}
	free((void *)s);
}

/**
 * \brief Returns current error value.
 *
 * \param s strbuf to extract the last error from.
 * \return the code of the last error.
 */
enum strbuf_errno_t
strbuf_errno_get(struct strbuf *s)
{
	assert(s != NULL);

	return (s->lasterr);
}


/**
 * \brief Clears current error value.
 *
 * \param s strbuf to act upon.
 */
void
strbuf_errno_clear(struct strbuf *s)
{
	assert(s != NULL);

	s->lasterr = strbuf_e_ok;
}


/**
 * \brief Translates the stored lasterr value to string.
 *
 * \param s strbuf to get the error from.
 * \return the static string that describes the error.
 */
const char*
strbuf_errno_to_s(struct strbuf *s)
{
	assert(s != NULL);

	switch (s->lasterr) {
		case strbuf_e_ok:
			return "everything is fine";
		case strbuf_e_nomem:
			return "not enough memory";
		case strbuf_e_cantgrow:
			return "can't grow buffer: already at maximal size";
		case strbuf_e_eof:
			return "reached end of file";
		default:
			return "OOPS: unknown strbuf errno code";
	}
	/* NOTREACHED */
}


/**
 * \brief Sets the current file descriptor
 * from which to read lines from.
 *
 * \param s strbuf to use.
 * \param fd file descriptor to read the line from.
 */
void
strbuf_set_fd(struct strbuf *s, int fd)
{
	s->fd = fd;
	s->ra_pos = NULL;
	s->ra_unread = 0;
}


/**
 * \brief Reads a single line from the given file descriptor.
 * The trailing '\n', if any, is preserved.
 *
 * \param s strbuf to use.
 * \return NULL if an error occurred or the read string
 * with terminating newline character.  strbuf's errno
 * is set and if its value is equal to strbuf_e_eof, then
 * we faced EOF before any characters were read.
 */
char *
strbuf_readline(struct strbuf *s)
{
	int fd;
	char c = '\0';
	ssize_t res = 0;
	size_t count = 0;

	assert(s != NULL);
	assert(s->fd != -1);

	fd = s->fd;

	_strbuf_make_empty(s);

	/*
	 * Assuming that there is something to read in the
	 * readahead buffer; if not, the block with
	 * s->ra_unread == 0 will correct the value of 'res'.
	 */
	res = 1;
	do {
		if (s->ra_unread == 0) {
			res = _strbuf_fill_buffer(s);
			if (res > 1)
				res = 1;
		}
		switch (res) {
			case 1:
				count++;
				c = *(s->ra_pos);
				s->ra_pos++;
				s->ra_unread--;
				_strbuf_add_char(s, c);
				break;
			case 0:
				if (count == 0) {
					s->lasterr = strbuf_e_eof;
					return NULL;
				}
				c = '\n';
				break;
			case -1:
				s->lasterr = strbuf_e_read_error;
				_strbuf_make_empty(s);
				return NULL;
		}
	} while (c != '\n');

	if (_strbuf_finalize_string(s)) {
		_strbuf_make_empty(s);
		return NULL;
	}

	return s->buf;
}

/**
 * \brief Returns the currently-stored string.
 *
 * \param s strbuf to get the string from.
 * \return the pointer to the stored string.
 */
char *
strbuf_get_string(struct strbuf *s)
{
	assert(s != NULL);

	return (s->buf);
}



/************************************************************************
 * Private functions.                                                   *
 ************************************************************************/

/*
 * Grows buffer for the given strbuf.
 *
 * Simple exponential grow: each time the size is doubled.
 */
static int
_strbuf_grow(struct strbuf *s)
{
	char *nbuf = NULL;
	size_t nsize = 0;

	nsize = (s->size >= SIZE_T_MAX/2 ? SIZE_T_MAX : s->size << 1);
	if (s->size >= nsize) {
		errno = strbuf_e_cantgrow;
		return -1;
	}

	nbuf = (char *)realloc(s->buf, nsize);
	if (nbuf == NULL) {
		errno = strbuf_e_nomem;
		return -1;
	}

	s->buf = nbuf;

	return 0;
}


/*
 * Creates empty string in the buffer.
 */
static void
_strbuf_make_empty(struct strbuf *s)
{
	s->offset = 0;
	s->buf[0] = '\0';
}


/*
 * Prepares the string in the buffer to be given away.
 *
 * Returns 0 if the addition was successful and -1 for error.
 */
static int
_strbuf_finalize_string(struct strbuf *s)
{
	return _strbuf_add_char(s, '\0');
}


/*
 * Adds one more character to the string buffer.
 *
 * Returns 0 if the addition was successful and -1 for error.
 */
static int
_strbuf_add_char(struct strbuf *s, char c)
{
	if (s->offset >= s->size) {
		int res = _strbuf_grow(s);
		if (res != 0)
			return res;
	}

	s->buf[s->offset] = c;
	s->offset++;

	return 0;
}

/*
 * Reads the next chunk from file to the readahead buffer.
 *
 * \param s string buffer to use.
 *
 * \return number of characters read into the buffer,
 * 0 for reaching end-of-file and -1 for any errors.
 */
static ssize_t
_strbuf_fill_buffer(struct strbuf *s)
{
	ssize_t res = read(s->fd, s->readahead, READAHEAD_SIZE);

	if (res <= 0) {
		s->ra_pos = NULL;
		s->ra_unread = 0;
		return res;
	}

	s->ra_pos = s->readahead;
	s->ra_unread = res;

	return res;
}
