#!/bin/sh
# Slow implementation of permissions adder.  Purely shell-based.

# Helper function that prints the octal mode for the given filesystem
# object to the standard output.  The output consists of 4 octal digits:
# the ordinary permissions and sticky/setXid bits.
#
# The object must exist and stat() call should be successful,
# otherwise function will not print anything.
getmod () {
	local mode

	if test -z "$1"; then
		echo "getmod(): called without arguments" > /dev/stderr
		exit 255
	fi

	if ! test -e "$1"; then
		return
	fi

	mode=`stat -f '%Mp%Lp' "$1" 2>/dev/null`
	if test "$?" != "0"; then
		return
	fi
	echo "$mode"
}

# Helper function that outputs the owner and the group for the
# filesystem object.  Textual forms are preferred to the numerical
# ones, but if it is not possible to convert UID/GID values to
# the user/group name, numericals will be used.
#
# Function outputs its data in the format 'user:group'.  In the case
# of error output will be empty.
getown () {
	local user group uname gname

	if [ -z "$1" ]; then
		echo "getown(): called without arguments" > /dev/stderr
		exit 255
	fi

	if ! [ -e "$1" ]; then
		return
	fi

	user=`stat -f '%Su' "$1"` 2>/dev/null
	if [ "$?" != "0" ]; then
		return
	fi
	group=`stat -f '%Sg' "$1"` 2>/dev/null
	if [ "$?" != "0" ]; then
		return
	fi

	echo "$user:$group"
}

# Adds trailing slash if it is missing.
#
# One argument: string.  The result is printed to the standard output.
add_trailing_slash () {
	if [ -z "$1" ]; then
		echo "add_trailing_slash(): called without arguments" > /dev/stderr
		exit 255
	fi

	echo "$1" | awk '{ if(substr($1, length($1), 1) == "/") print $0; else print $0 "/"; }'
}

# Removes trailing shash if it is here.
#
# One argument: string.  The result is printed to the standard output.
remove_trailing_slash () {
	if [ -z "$1" ]; then
		echo "remove_trailing_slash(): called without arguments" > /dev/stderr
		exit 255
	fi

	echo "$1" | sed -e's|/$||'
}

# Reads the file list from the standard input and appends
# octal file mode and ownership to the file specifications.
# Other lines, including comments are left intact.
# The produced lines are written to the standard output.
#
# If the specified filesystem object does not exists, the
# corresponding line will be commented out with the verbose
# explanation why is was commented.
#
# Expects one argument: list mode.
# - 'all' produces the list with the actual permissions,
# - 'missing' only adds permissions to the lines with missing
#   or insane permissions field.
filelist_add_permissions () {
	local line file mode ownership

	if [ -z "$1" ]; then
		echo "filelist_add_permissions(): called without arguments" > /dev/stderr
		exit 255
	fi

	case "$1" in
	all|missing)
		;;
	*)
		echo "filelist_add_permissions(): first argument has bad value '$1'" > /dev/stderr
		exit 255
		;;
	esac

	while read -r line; do
		case "X$line" in
		X|X#*)
			echo "$line"
			continue
			;;
		esac
		set -- $line
		file="$1"
		mode="$2"
		ownership="$3"

		# Comment-out nonexistent filesystem objects.
		if [ ! -e "$file" ]; then
			echo "#NONEXISTENT: $line"
			continue
		fi

		#
		# Check that mode and ownership are more-or-less
		# sane: we're reusing the values from the
		# configuration file.  If they are bad, we will
		# use the values from filesystem.
		#
		case "$mode" in
		[0-7][0-7][0-7])
			;;
		*)
			mode=''
			;;
		esac
		case "$ownership" in
		*:*)
			;;
		*)
			ownership=''
			;;
		esac

		if [ -d "$file" -a ! -L "$file" ]; then
			file=`add_trailing_slash "$file"`
		fi

		if [ -z "$mode" -o -z "$ownership" -o "$1" != "missing" ];
		then

			if [ -z "$mode" ]; then
				mode=`getmod "$file"`
				if [ -z "$mode" ]; then
					mode=unknown
				fi
			fi
			if [ -z "$ownership" ]; then
				ownership=`getown "$file"`
				if [ -z "$ownership" ]; then
					ownership=unknown
				fi
			fi
			printf "%-50s  %s  %s\n" "$file" \
			    "$mode" "$ownership"
		else
			echo "$line"
		fi
	done
}

ME=$(basename "$0")
if [ -z "$ME" ]; then
	echo "Can't determine my own name." >&2
	exit 1
fi

IN=$(mktemp -t "$ME".in) || exit 1
OUT1=$(mktemp -t "$ME".out1) || exit 1
OUT2=$(mktemp -t "$ME".out2) || exit 1

find /etc/ > "$IN" 2>/dev/null
filelist_add_permissions all < "$IN" > "$OUT1" || exit 1
../cft-add-permissions all < "$IN" > "$OUT2" || exit 1
cmp "$OUT1" "$OUT2" >&2
