/**
 * \brief Implementation of non-failing *alloc() and friends.
 *
 * These functions are to be used in place of the ordinary
 * malloc, realloc and free when the caller will not be able
 * to operate without the memory being allocated, reallocated
 * or freed.
 */

#include <stdlib.h>

#include "err.h"
#include "sysexits.h"
#include "xalloc.h"

/**
 * \brief Allocates the requested amount of memory.
 *
 * \param size number of bytes to allocate.
 * \return the pointer to the allocated memory.
 */
void *
xmalloc(size_t size)
{
	void *retval = malloc(size);

	if (retval != NULL)
		return retval;

	errx(EX_SOFTWARE, "Can't allocate %zd bytes of memory.",
	  size);
}


/**
 * \brief Allocates the memory for the given number of
 * objects each of a specified size.
 *
 * \param number number of objects.
 * \param size size of a single object.
 *
 * \return The pointer to the newly-allocated area.
 */
void *
xcalloc(size_t number, size_t size)
{
	return xmalloc(number * size);
}


/**
 * \brief Reallocates the given chunk of memory (previously
 * allocated with xmalloc) to the given size.
 *
 * \param ptr pointer to the memory block to reallocate.
 * \param size new size.
 * \return pointer to the reallocated block.
 */
void *
xrealloc(void *ptr, size_t size)
{
	void *retval = realloc(ptr, size);

	if (retval != NULL)
		return retval;

	errx(EX_SOFTWARE, "Can't reallocate memory chunk to %zd bytes.",
	  size);
}


/**
 * \brief Frees previously allocated memory block.
 *
 * \param ptr pointer to the memory block to free.
 */
void
xfree(void *ptr)
{
	free(ptr);
}
